
#ifndef MP_FILE_WRAPPER_H_
#define MP_FILE_WRAPPER_H_

#include <string>

#ifdef _MSC_VER
//#include <stdlib.h>  // This is used only for mbstowcs_s() .
#include <algorithm> // This is used only for std::transform() .
#include <cctype>    // This is used only for std::toupper() .
#else
#include <sys/stat.h>
#endif


#include "stdio_synchronizer.hpp"

namespace multiplatform {
namespace file {

// By default, no error messages are displayed even if an empty string is entered.
// However, since the file naturally does not exist, the return value is false.
// This behavior was done to prevent excessive error messages from being displayed
// of the empty lines when reading the file list.
// If you want to need display a message that is an empty string when using it
// in processing of a single file, please set it to "ignore_blank_msg = false".
inline bool file_exists(std::string filename, bool ignore_blank_msg = true){
  std::string::size_type n = filename.size();
#ifdef _MSC_VER
  if(n > 1 && (filename[n-1] == '\\' || filename[n-1] == '/') && !(n == 3 && filename[1] == ':')){ filename.resize(n-1); }
  if(filename.empty()){
    if(!ignore_blank_msg){ _warning_print("The file name is an empty string."); }
    return false;
  }
  //size_t wcs_size;
  //wchar_t *wcs = new wchar_t[filename.length() + 1];
  //mbstowcs_s(&wcs_size, wcs, filename.length() + 1, filename.c_str(), filename.length() + 1);
  //DWORD attr = GetFileAttributesW((LPCWSTR)wcs);
  //delete[] wcs;
  std::wstring wstr(filename.begin(), filename.end());
  DWORD attr = GetFileAttributesW((LPCWSTR)wstr.c_str());
  // is_not_exsits
  if(attr == INVALID_FILE_ATTRIBUTES){
    std::string msg; msg.reserve(filename.size() + 100);
    _warning_print(msg.append("Cannot read file: ").append(filename).append(" (reason: file is not exists)"));
    return false;
  }
  // is_directory
  if((attr & FILE_ATTRIBUTE_DIRECTORY) != 0){
    std::string msg; msg.reserve(filename.size() + 100);
    _warning_print(msg.append("Cannot read file: ").append(filename).append(" (reason: this is a directory)"));
    return false;
  }
  // is_device COM or LPT of larger number than 10 ... e.g.) '\\.\COM13'
  if(filename.compare(0, 4, "\\\\.\\") == 0){
    std::string msg; msg.reserve(filename.size() + 100);
    _warning_print(msg.append("Cannot read file: ").append(filename).append(" (reason: this is a device)"));
    return false;
  }
  // is_device
  {
    std::string ifilename = filename;
    std::transform(ifilename.begin(), ifilename.end(), ifilename.begin(), std::toupper);
    n = ifilename.size();
    // is_device ... e.g.) 'CON', 'PRN', 'AUX' or 'NUL'
    if(n == 3 && (ifilename.compare("CON") == 0 ||
                  ifilename.compare("PRN") == 0 ||
                  ifilename.compare("AUX") == 0 ||
                  ifilename.compare("NUL") == 0)){
      std::string msg; msg.reserve(filename.size() + 100);
      _warning_print(msg.append("Cannot read file: ").append(filename).append(" (reason: this is a device)"));
      return false;
    }
    // is_device ... e.g.) 'COM0' to 'COM9' or 'LPT0' to 'LPT9'
    if(n == 4 && (ifilename.compare(0, 3, "COM") == 0 ||
                  ifilename.compare(0, 3, "LPT") == 0)){
      if(ifilename[3] >= 0x30 && ifilename[3] <= 0x39){
        std::string msg; msg.reserve(filename.size() + 100);
        _warning_print(msg.append("Cannot read file: ").append(filename).append(" (reason: this is a device)"));
        return false;
      }
    }
    // is_device with path
    if(n > 3){
      bool found_keyword = false;
      std::string::size_type index = 0;
      while(index != std::string::npos){
        index = ifilename.find("\\", index);
        if(index != std::string::npos){
          // is_device ... e.g.) 'C:\CON', '.\anything\PRN', '..\aux\dummy.doc' or 'nul.htlm'
          if(index+3 < n){
            if(ifilename.compare(index+1, 3, "CON") == 0 ||
               ifilename.compare(index+1, 3, "PRN") == 0 ||
               ifilename.compare(index+1, 3, "AUX") == 0 ||
               ifilename.compare(index+1, 3, "NUL") == 0){
              if(index+4 < n){
                if(ifilename.compare(index+4, 1, ".") == 0){
                  found_keyword = true;
                  break; //index = std::string::npos;
                }
              }
              else{
                found_keyword = true;
                break; //index = std::string::npos;
              }
            }
          }
          // is_device ... e.g.) 'C:\dummy\com1' or 'LPT2.csv'
          if(index+4 < n){
            if(ifilename.compare(index+1, 3, "COM") == 0 ||
               ifilename.compare(index+1, 3, "LPT") == 0){
              if(ifilename[index+4] >= 0x30 && ifilename[index+4] <= 0x39){
                if(index+5 < n){
                  if(ifilename.compare(index+5, 1, ".") == 0){
                    found_keyword = true;
                    break; //index = std::string::npos;
                  }
                }
                else{
                  found_keyword = true;
                  break; //index = std::string::npos;
                }
              }
            }
          }
          if(index+1 < n && index != std::string::npos){ index++; }
        }
      }
      if(found_keyword){
        std::string msg; msg.reserve(filename.size() + 100);
        _warning_print(msg.append("Cannot read file: ").append(filename).append(" (reason: this is a device)"));
        return false;
      }
    }
  }
#else // _MSC_VER
  if(n > 1 && filename[n-1] == '/'){ filename.resize(n-1); }
  if(filename.empty()){
    if(!ignore_blank_msg){ _warning_print("The file name is an empty string."); }
    return false;
  }
  {
    struct stat st;
    if(stat(filename.c_str(), &st) != 0){
      std::string msg; msg.reserve(filename.size() + 100);
      _warning_print(msg.append("Cannot read file: ").append(filename).append(" (reason: stat() failed)"));
      return false;
    }
    if(!S_ISREG(st.st_mode)){
      std::string msg; msg.reserve(filename.size() + 100);
      _warning_print(msg.append("Cannot read file: ").append(filename).append(" (reason: file is not regular file)"));
      return false;
    }
  }
#endif // _MSC_VER
  return true;
}

} // namespace file
} // namespace multiplatform

#endif // MP_SLEEP_WRAPPER_H_

