
#ifndef MP_MFCNET_SERVER_H_
#define MP_MFCNET_SERVER_H_

#include "hark-support.h"
#include "stdio_synchronizer.hpp"
#include "io_manager.hpp"
#include "tcp_server.hpp"
#include "tcp_socket.hpp"
#include "thread_pool.hpp"
#include "wsa_wrapper.hpp"

#include <atomic>
#include <sstream>
#include <mutex>
#include <map>
#include <unordered_map>
#include <deque>
#include <utility>
#include <functional>
#include <cstdint>
#include <iterator>

class mfcnet_data {
public:
  mfcnet_data():
    m_srcinfo_ready(false), m_srcdata_ready(false), m_completed(false),
    m_offset(0), m_dim(0), m_feature_bytes(0), m_frames(0)
  {
    m_srcinfo.source_id = 0;
    m_srcinfo.azimuth = 0.0f;
    m_srcinfo.elevation = 0.0f;
    m_srcinfo.tv_sec = 0;
    m_srcinfo.tv_usec = 0;
  }

  ~mfcnet_data(){
  }

  //std::unique_lock<std::mutex> get_lock(){
  //  return std::unique_lock<std::mutex>(m_buffer_mutex);
  //}
  //void push_data(std::vector<char> data){
  //  std::unique_lock<std::mutex> lock = get_lock();
  //  std::copy(data.begin(), data.end(), std::back_inserter(m_buffer));
  //}

  void push_data(std::vector<char> data){
    std::lock_guard<std::mutex> lock(m_buffer_mutex);
    std::copy(data.begin(), data.end(), std::back_inserter(m_buffer));
  }

  inline bool srcinfo_is_ready(){ return m_srcinfo_ready; }
  inline bool srcdata_is_ready(){ return m_srcdata_ready; }
  inline bool is_completed(){ return m_completed; }

  inline std::int32_t get_feature_bytes(){ return m_feature_bytes; }
  inline std::int32_t get_feature_dimension(){ return m_dim; }
  inline std::int32_t get_available_frames(){ return m_frames; }

  struct SourceInfo get_source(){
    // Data race does not occur even if reading occurs at the same time.
    if(srcinfo_is_ready()){ return m_srcinfo; }
    if(check_buffer(sizeof(std::int32_t) + sizeof(struct SourceInfo))){
      std::int32_t srcinfo_size = pop_int32_t(get_offset());
      if(srcinfo_size != sizeof(struct SourceInfo)){
        {
          std::stringstream ss;
          ss << "header size missmatch: " << srcinfo_size << " vs " << sizeof(struct SourceInfo);
          _warning_print(ss.str());
        }
        {
          struct SourceInfo empty_srcinfo;
          empty_srcinfo.source_id = 0;
          empty_srcinfo.azimuth = 0.0f;
          empty_srcinfo.elevation = 0.0f;
          empty_srcinfo.tv_sec = 0;
          empty_srcinfo.tv_usec = 0;
          return empty_srcinfo;
        }
      }
      add_offset(sizeof(std::int32_t));
      // Since it is protected by the m_srinfo_ready flag,
      // writing and reading do not occur at the same time.
      m_srcinfo = pop_SourceInfo(get_offset());
      add_offset(sizeof(struct SourceInfo));
      //DEBUG//{
      //DEBUG//  std::stringstream ss;
      //DEBUG//  ss << "source_id = "   << m_srcinfo.source_id
      //DEBUG//     << ", azimuth = "   << m_srcinfo.azimuth
      //DEBUG//     << ", elevation = " << m_srcinfo.elevation
      //DEBUG//     << ", sec = "       << m_srcinfo.tv_sec
      //DEBUG//     << ", usec = "      << m_srcinfo.tv_usec;
      //DEBUG//  _raw_print_with_endl(ss.str());
      //DEBUG//}
      m_srcinfo_ready = true; // atomic
      return m_srcinfo;
    }
    // Reaching here means that data has not accumulated in the buffer yet.
    {
      struct SourceInfo empty_srcinfo;
      empty_srcinfo.source_id = 0;
      empty_srcinfo.azimuth = 0.0f;
      empty_srcinfo.elevation = 0.0f;
      empty_srcinfo.tv_sec = 0;
      empty_srcinfo.tv_usec = 0;
      return empty_srcinfo;
    }
  }

  std::vector<float> get_feature(){
    // The sound source information is not ready means that we have not received feature data yet.
    if(!srcinfo_is_ready()){
      std::vector<float> empty_srcdata;
      return empty_srcdata;
    }
    do{
      // If completed, the feature data receive processing is immediately terminated.
      if(is_completed()){ break; }

      // Since data to be processed is not accumulated yet, it exits from the method.
      if(!check_buffer(get_offset() + sizeof(std::int32_t))){ break; }

      std::int32_t feature_bytes = pop_int32_t(get_offset());

      // We will only perform termination checks. We do not update m_offset yet at this point.
      if(feature_bytes == 0){
        m_completed = true; // atomic
        break;
      }

      // Updated only when the dimension number of the received feature quantity is not set.
      if(get_feature_bytes() == 0){
        m_feature_bytes = feature_bytes; // atomic
        m_dim = (feature_bytes / sizeof(float)); // atomic
      }

      // Since abnormality was detected, processing is aborted.
      // Socket connection and reception processing will continue, but will not be used.
      // We ignore the data after the error detection because we have detected that
      // the number of dimensions of feature has changed or that we received incorrect data.
      if(feature_bytes != get_feature_bytes()){
        {
          std::stringstream ss;
          ss << "source_id = " << m_srcinfo.source_id << ", feature_bytes = " << feature_bytes << " (!=" << get_feature_bytes() << ")" << std::endl;
          _warning_print(ss.str());
        }
        m_completed = true; // atomic
        break;
      }

      // Since data to be processed is not accumulated yet, it exits from the method.
      if(!check_buffer(get_offset() + (sizeof(std::int32_t) + feature_bytes) * 2)){ break; }

      add_offset(sizeof(std::int32_t));
      std::vector<float> feature = pop_vec_float(get_offset(), m_dim);
      add_offset(feature_bytes);

      std::int32_t mask_bytes = pop_int32_t(get_offset());

      // Since abnormality was detected, processing is aborted.
      // Socket connection and reception processing will continue, but will not be used.
      // We ignore the data after the error detection because we have detected that
      // the number of dimensions of mask has changed or that we received incorrect data.
      if(mask_bytes != get_feature_bytes()){
        {
          std::stringstream ss;
          ss << "source_id = " << m_srcinfo.source_id << ", mask_bytes = " << mask_bytes << " (!=" << get_feature_bytes() << ")" <<  std::endl;
          _warning_print(ss.str());
        }
        m_completed = true; // atomic
        break;
      }

      add_offset(sizeof(std::int32_t));
      // Since Kaldi does not support the missing feature mask function, it does not process it.
      //std::vector<float> mask = pop_vec_float(get_offset(), m_dim);
      add_offset(mask_bytes);

      std::copy(feature.begin(), feature.end(), std::back_inserter(m_srcdata));
      m_frames++; // atomic
      m_srcdata_ready = true; // atomic

      //DEBUG//{
      //DEBUG//  std::stringstream ss;
      //DEBUG//  ss << "m_dim = " << m_dim << ", m_offset = " << get_offset();
      //DEBUG//  _raw_print_with_endl(ss.str());
      //DEBUG//}
    } while(true);
    //DEBUG//{
    //DEBUG//  std::stringstream ss;
    //DEBUG//  ss << "source_id = " << m_srcinfo.source_id << ", received frames = " << get_available_frames() << std::endl;
    //DEBUG//  _raw_print_with_endl(ss.str());
    //DEBUG//}
    return m_srcdata;
  }
//##### Why is not protected by mutex and lock against m_srcdata? This is a simple trick. #####//
////
// Since it is known that the user who received m_srcdata can read up to m_frames,
// it is guaranteed not to read to elements (> m_frames) that may be written.
// Also, data race does not occur even if reading occurs at the same time.
////
// There are definitions and examples in "23.2.2 Container data races / p2-3" of N3290 as follows.
// -----------------------------------------------------------------------------------------------
// 2. Notwithstanding (17.6.5.9),
//   implementations are required to avoid data races when the contents of the contained object
//   in different elements in the same sequence, excepting vector<bool>, are modified concurrently.
// 3. [ Note: For a vector<int> x with a size greater than one, x[1] = 5 and *x.begin() = 10
//   can be executed concurrently without a data race, but x[0] = 5 and *x.begin() = 10 executed
//   concurrently may result in a data race. As an exception to the general rule,
//   for a vector<bool> y, y[0] = true may race with y[1] = true. -- end note ]
// -----------------------------------------------------------------------------------------------
////
// *) In other words, except for specialized std::vector <bool>, different elements of the same
//   sequence container can be safely accessed from another thread without causing a data race.

private:
  inline std::int32_t get_offset(){  std::lock_guard<std::mutex> lock(m_offset_mutex); return m_offset; }
  inline void add_offset(std::int32_t value){ std::lock_guard<std::mutex> lock(m_offset_mutex); m_offset += value; }
  inline void set_offset(std::int32_t offset){ std::lock_guard<std::mutex> lock(m_offset_mutex); m_offset = offset; }
  inline void reset_offset(){ std::lock_guard<std::mutex> lock(m_offset_mutex); m_offset = 0; }

  inline bool check_buffer(std::size_t size){
    std::lock_guard<std::mutex> lock(m_buffer_mutex);
    return (m_buffer.size() >= size) ? true : false;
  }

  inline std::int32_t pop_int32_t(std::int32_t offset){
    union U{
      std::int32_t    i;
      char            c[sizeof(std::int32_t)];
    } u;
    {
      std::lock_guard<std::mutex> lock(m_buffer_mutex);
      std::memcpy(u.c, reinterpret_cast<char*>(m_buffer.data())+offset, sizeof(std::int32_t));
    }
    return u.i;
  }

  inline struct SourceInfo pop_SourceInfo(std::int32_t offset){
    union U{
      struct SourceInfo    s;
      char                 c[sizeof(struct SourceInfo)];
    } u;
    {
      std::lock_guard<std::mutex> lock(m_buffer_mutex);
      std::memcpy(u.c, reinterpret_cast<char*>(m_buffer.data())+offset, sizeof(struct SourceInfo));
    }
    return u.s;
  }

  inline std::vector<float> pop_vec_float(std::int32_t offset, std::int32_t count){
    union U{
      float    f;
      char     c[sizeof(float)];
    } u;
    std::vector<float> vec; vec.resize(count);
    {
      std::lock_guard<std::mutex> lock(m_buffer_mutex);
      char *p = reinterpret_cast<char*>(m_buffer.data())+offset;
      for(int i=0; i<count; i++, p+=sizeof(float)){
        std::memcpy(u.c, p, sizeof(float));
        vec[i] = u.f;
      }
    }
    return vec;
  }

  // Flags
  std::atomic<bool>         m_srcinfo_ready; // The sound source information header is ready to be read.
  std::atomic<bool>         m_srcdata_ready; // The feature data of the sound source is ready to be read.
  std::atomic<bool>         m_completed;     // The data will not increase.

  std::vector<char>         m_buffer;        // input buffer
  std::int32_t              m_offset;        // proccessed bytes

  struct SourceInfo         m_srcinfo;       // source info
  std::vector<float>        m_srcdata;       // source features

  std::atomic<std::int32_t> m_dim;           // The dimension of the features.
  std::atomic<std::int32_t> m_feature_bytes; // The size of the features.
  std::atomic<std::int32_t> m_frames;        // The number of currently available frames.

  std::mutex                m_buffer_mutex;  // for m_buffer
  std::mutex                m_offset_mutex;  // for m_offset,
};

class mfcnet_server {
public:
  mfcnet_server(){
  }

  mfcnet_server(const std::string& host, std::uint32_t port){
    s.start(host, port, [&](const std::shared_ptr<multiplatform::tcp_client>& client) -> bool {
      {
        std::stringstream ss;
        ss << "Client connected: " << client->get_host() << ":" << client->get_port();
        _raw_print_with_endl(ss.str());
      }
      buffers[client->get_port()] = new mfcnet_data();
      client->async_read({1024, std::bind(&mfcnet_server::on_new_message, this, client, std::placeholders::_1)});
      {
        std::lock_guard<std::mutex> lock(m_nmap_mutex);
        nmap.insert(std::pair<std::uint32_t, bool>(client->get_port(), true));
      }
      return true;
    });
  }

  ~mfcnet_server(){
    for(auto buffer : buffers){
      if(buffer.second != nullptr){
        delete buffer.second;
      }
    }
  }

  bool erase_buffer_key(std::uint32_t& port){
    if(buffers.empty()){ return false; }
    bool ret = false;
    {
      if(buffers.find(port) != buffers.end() && buffers[port] != nullptr){
        delete buffers[port];
        buffers[port] = nullptr;
      }
      if(buffers.erase(port) != 0){ ret = true; }
    }
    return ret;
  }

  std::size_t data_read(){
    std::vector<float> srcdata;
    for(auto buffer : buffers){
      srcdata = buffer.second->get_feature();
    }
    return 0;
  }

  std::size_t src_read(){
    struct SourceInfo srcinfo;
    for(auto buffer : buffers){
      srcinfo = buffer.second->get_source();
    }
    return 0;
  }

  bool find_source(std::int32_t src_id, struct SourceInfo &ret){
    struct SourceInfo srcinfo;
    for(auto buffer : buffers){
      srcinfo = buffer.second->get_source();
      if(srcinfo.source_id == src_id){
        ret = srcinfo;
        return true;
      }
    }
    return false;
  }

  bool find_feature(std::int32_t src_id, std::vector<float> &ret, bool completed){
    struct SourceInfo srcinfo;
    for(auto buffer : buffers){
      srcinfo = buffer.second->get_source();
      if(srcinfo.source_id == src_id){
        ret = buffer.second->get_feature();
        return completed ? buffer.second->is_completed() : true;
      }
    }
    return false;
  }

  std::deque<std::uint32_t> get_new_keys(bool completed = false){
    std::deque<std::uint32_t> tmp;
    if(completed){
      std::lock_guard<std::mutex> lock(m_nmap_mutex);
      for(auto n : nmap){
        if(buffers[n.first]->is_completed()){
          tmp.push_back(n.first);
        }
      }
    }
    else{
      std::lock_guard<std::mutex> lock(m_nmap_mutex);
      for(auto n : nmap){ tmp.push_back(n.first); }
    }
    return tmp;
  }

  bool erase_new_key(std::uint32_t& port){
    if(nmap.empty()){ return false; }
    bool ret = false;
    {
      std::lock_guard<std::mutex> lock(m_nmap_mutex);
      if(nmap.erase(port) != 0){ ret = true; }
    }
    return ret;
  }

  bool find_new_key(std::uint32_t& port){
    if(nmap.empty()){ return false; }
    bool ret = false;
    {
      std::lock_guard<std::mutex> lock(m_nmap_mutex);
      if(nmap.find(port) != nmap.end()){ ret = true; }
    }
    return ret;
  }

  std::deque<std::uint32_t> get_terminated_keys(bool completed = false){
    std::deque<std::uint32_t> tmp;
    if(completed){
      std::lock_guard<std::mutex> lock(m_dmap_mutex);
      for(auto d : dmap){
        if(buffers[d.first]->is_completed()){
          tmp.push_back(d.first);
        }
      }
    }
    else{
      std::lock_guard<std::mutex> lock(m_dmap_mutex);
      for(auto d : dmap){
        tmp.push_back(d.first);
      }
    }
    return tmp;
  }

  bool erase_terminated_key(std::uint32_t& port){
    if(dmap.empty()){ return false; }
    bool ret = false;
    {
      std::lock_guard<std::mutex> lock(m_dmap_mutex);
      if(dmap.erase(port) != 0){ ret = true; }
    }
    return ret;
  }

  bool find_terminated_key(std::uint32_t& port){
    if(dmap.empty()){ return false; }
    bool ret = false;
    {
      std::lock_guard<std::mutex> lock(m_dmap_mutex);
      if(dmap.find(port) != dmap.end()){ ret = true; }
    }
    return ret;
  }

  void on_new_message(const std::shared_ptr<multiplatform::tcp_client>& client, multiplatform::tcp_client::read_response& res) {
    if(res.success){
      //DEBUG//{
      //DEBUG//  std::stringstream ss;
      //DEBUG//  ss << "Client data received: " << client->get_host() << ":" << client->get_port();
      //DEBUG//  _raw_print_with_endl(ss.str());
      //DEBUG//}
      buffers[client->get_port()]->push_data(res.buffer);
      //DEBUG// std::string recv_msg(res.buffer.begin(), res.buffer.end()); _raw_print_with_endl(recv_msg); // print recv message
      client->async_read({1024, std::bind(&mfcnet_server::on_new_message, this, client, std::placeholders::_1)});
    }
    else {
      {
        std::stringstream ss;
        ss << "Client disconnected: " << client->get_host() << ":" << client->get_port();
        _raw_print_with_endl(ss.str());
      }
      {
        std::lock_guard<std::mutex> lock(m_dmap_mutex);
        dmap.insert(std::pair<std::uint32_t, bool>(client->get_port(), true));
      }
      client->disconnect();
    }
  }

  multiplatform::tcp_server s;
  std::map<std::uint32_t, mfcnet_data* > buffers; // key:port, data:buffer
  std::unordered_map<std::uint32_t, bool> nmap; // new connected port map (new sources)
  std::unordered_map<std::uint32_t, bool> dmap; // disconnected port map (terminated sources)
  std::mutex m_nmap_mutex;
  std::mutex m_dmap_mutex;
};

#endif // MP_MFCNET_SERVER_H_

