
#ifndef MP_TCP_CLIENT_H_
#define MP_TCP_CLIENT_H_

#include "stdio_synchronizer.hpp"
#include "io_manager.hpp"
#include "tcp_socket.hpp"

#include <atomic>
#include <condition_variable>
#include <functional>
#include <cstdint>
#include <list>
#include <memory>
#include <mutex>
#include <string>
#include <vector>
#include <queue>
#include <thread>
#include <unordered_map>

#ifdef _WIN32
#include <Winsock2.h>
#else
#include <sys/select.h>
#include <unistd.h>
#endif /* _WIN32 */

#include <cstdint>
#include <stdexcept>
#include <iostream>
#include <string>
#include <memory>
#include <mutex>

// Note:
//   If it is using auto type, it will be easier to read because #ifdef lines are reduced.
//   #ifdef was caused by the difference of socket file descriptor type on Windows and Ubuntu.

namespace multiplatform {

// multiplatform::tcp_server is the class providing TCP Client features.
// The tcp_client works entirely asynchronously
class tcp_client {
public:

#ifdef _WIN32
  using socket_fd_t = SOCKET;
#else // _WIN32
  using socket_fd_t = int;
#endif // _WIN32

  tcp_client(void): m_disconnection_handler(nullptr){
    m_io_manager = get_default_io_manager();
    if(m_debug){ _debug_print("create tcp_client"); }
  }
  ~tcp_client(void){
    if(m_debug){ _debug_print("destroy tcp_client"); }
    disconnect(true);
  }
  tcp_client(const tcp_client&) = delete;            // Removed copy constructor
  tcp_client& operator=(const tcp_client&) = delete; // Removed assignment operator

  // Build a client from an existing tcp_socket instance (the socket is moved)
  explicit tcp_client(tcp_socket&& socket):
      m_io_manager(get_default_io_manager()), m_socket(std::move(socket)), m_disconnection_handler(nullptr){
    m_is_connected = true;
    if(m_debug){ _debug_print("create tcp_client"); }
    m_io_manager->track(m_socket);
  }

  // Comparison operator
  bool operator==(const tcp_client& rhs) const{ return m_socket == rhs.m_socket; }
  bool operator!=(const tcp_client& rhs) const{ return m_socket != rhs.m_socket; }

  // Getter
  const std::string& get_host(void) const{ return m_socket.get_host(); }
  std::uint32_t get_port(void) const{ return m_socket.get_port(); }

  // Connect the socket to the remote server.
  //
  // \param host Hostname of the target server
  // \param port Port of the target server
  // \param timeout_msecs maximum time to connect (will block until connect succeed or timeout expire). 0 will block undefinitely. If timeout expires, connection fails
  void connect(const std::string& host, std::uint32_t port, std::uint32_t timeout_msecs = 0){
    if(is_connected()){
      _error_throw("tcp_client is already connected");
    }
    try{
      m_socket.connect(host, port, timeout_msecs);
      m_io_manager->track(m_socket);
    }
    catch(const std::runtime_error& e){
      m_socket.close();
      throw e;
    }
    m_is_connected = true;
    if(m_debug){ _info_print("tcp_client connected"); }
  }

  // Disconnect the tcp_client if it was currently connected.
  //
  // \param wait_for_removal When sets to true, disconnect blocks until the underlying TCP client has been effectively removed from the io_service and that all the underlying callbacks have completed.
  void disconnect(bool wait_for_removal = false){
    if(!is_connected()){ return; }
    // update state
    m_is_connected = false;
    // clear all pending requests
    clear_read_requests();
    clear_write_requests();
    // remove socket from io service and wait for removal if necessary
    m_io_manager->untrack(m_socket);
    if(wait_for_removal){ m_io_manager->wait_for_removal(m_socket); }
    // close the socket
    m_socket.close();
    if(m_debug){ _info_print("tcp_client disconnected"); }
  }

  // \return whether the client is currently connected or not
  bool is_connected(void) const{ return m_is_connected; }

  // Structure that stores responses to read operations.
  struct read_response {
    bool success;             // Whether the operation was successful. If false, the client is probably disconnected.
    std::vector<char> buffer; // Buffer of read data.
  };

  // Structure that stores responses to write operations.
  struct write_response {
    bool success;     // Whether the operation was successful. If false, the client is probably disconnected.
    std::size_t size; // Number of bytes written.
  };

  // callback to be called on async read completion
  // takes the read_response as a parameter
  using async_read_callback_t = std::function<void(read_response&)>;

  // callback to be called on async write completion
  // takes the write_response as a parameter
  using async_write_callback_t = std::function<void(write_response&)>;

  // Structure that stores read requests.
  // Note: Even if you read less than the number of bytes requested, the callback will be invoked.
  struct read_request {
    std::size_t size;                          // Number of bytes to read.
    async_read_callback_t async_read_callback; // Callback executed at the completion of the read operation.
  };

  // Structure that stores write requests.
  // Note: Even if you write less than the number of bytes requested, the callback will be invoked.
  struct write_request {
    std::vector<char> buffer;                    // Buffer of write data.
    async_write_callback_t async_write_callback; // Callback executed at the completion of the write operation.
  };

  // Asynchronous read operation
  void async_read(const read_request& request){
    std::lock_guard<std::mutex> read_lock(m_read_requests_mutex);
    if(is_connected()){
      m_io_manager->set_rd_callback(m_socket, std::bind(&tcp_client::on_read_available, this, std::placeholders::_1));
      m_read_requests.push(request);
    }
    else{
      _error_throw("tcp_client is disconnected");
    }
  }

  // Asynchronous write operation
  void async_write(const write_request& request){
    std::lock_guard<std::mutex> write_lock(m_write_requests_mutex);
    if(is_connected()){
      m_io_manager->set_wr_callback(m_socket, std::bind(&tcp_client::on_write_available, this, std::placeholders::_1));
      m_write_requests.push(request);
    }
    else{
      _error_throw("tcp_client is disconnected");
    }
  }

  // tcp_socket getter
  multiplatform::tcp_socket& get_socket(void){ return m_socket; }
  const multiplatform::tcp_socket& get_socket(void) const{ return m_socket; }

  // \return io service monitoring this tcp connection
  const std::shared_ptr<multiplatform::io_manager>& get_io_manager(void) const{ return m_io_manager; }

  // disconnection handle
  // called whenever a disconnection occured
  using disconnection_handler_t = std::function<void()>;

  // set on disconnection handler
  //
  // \param disconnection_handler the handler to be called on disconnection
  void set_on_disconnection_handler(const disconnection_handler_t& disconnection_handler){ m_disconnection_handler = disconnection_handler; }

private:
  // Call the user-defined disconnection handler
  void call_disconnection_handler(void){
    if(m_disconnection_handler){ m_disconnection_handler(); }
  }

  // io service read callback
  // called by the io service whenever the socket is readable
  //
  // \param fd file description of the socket for which the read is available
  void on_read_available(socket_fd_t fd){
    if(m_debug){ _info_print("read available"); }
    read_response res;
    auto callback = process_read(res);
    if(!res.success){
      _warning_print("read operation failure");
      disconnect();
    }
    if(callback){ callback(res); }
    if(!res.success){ call_disconnection_handler(); }
  }

  // io service write callback
  // called by the io service whenever the socket is writable
  //
  // \param fd file description of the socket for which the write is available
  void on_write_available(socket_fd_t fd){
    if(m_debug){ _info_print("write available"); }
    write_response res;
    auto callback = process_write(res);
    if(!res.success){
      _warning_print("write operation failure");
      disconnect();
    }
    if(callback){ callback(res); }
    if(!res.success){ call_disconnection_handler(); }
  }

  // Clear pending read requests (basically empty the queue of read requests)
  void clear_read_requests(void){
    std::lock_guard<std::mutex> read_lock(m_read_requests_mutex);
    std::queue<read_request> empty;
    std::swap(m_read_requests, empty);
  }

  // Clear pending write requests (basically empty the queue of write requests)
  void clear_write_requests(void){
    std::lock_guard<std::mutex> write_lock(m_write_requests_mutex);
    std::queue<write_request> empty;
    std::swap(m_write_requests, empty);
  }

  // process read operations when available
  // basically called whenever on_read_available is called and try to read from the socket
  // handle possible case of failure and fill in the result
  async_read_callback_t process_read(read_response& response){
    std::lock_guard<std::mutex> read_lock(m_read_requests_mutex);
    if(m_read_requests.empty()){ return nullptr; }
    const auto& request = m_read_requests.front();
    auto callback       = request.async_read_callback;
    try{
      response.buffer  = m_socket.recv(request.size);
      response.success = true;
    }
    catch(const std::runtime_error&){
      response.success = false;
    }
    m_read_requests.pop();
    if(m_read_requests.empty()){ m_io_manager->set_rd_callback(m_socket, nullptr); }
    return callback;
  }

  // process write operations when available
  // basically called whenever on_write_available is called and try to write to the socket
  // handle possible case of failure and fill in the result
  async_write_callback_t process_write(write_response& response){
    std::lock_guard<std::mutex> write_lock(m_write_requests_mutex);
    if(m_write_requests.empty()){ return nullptr; }
    const auto& request = m_write_requests.front();
    auto callback       = request.async_write_callback;
    try{
      response.size    = m_socket.send(request.buffer, request.buffer.size());
      response.success = true;
    }
    catch(const std::runtime_error&){
      response.success = false;
    }
    m_write_requests.pop();
    if(m_write_requests.empty()){ m_io_manager->set_wr_callback(m_socket, nullptr); }
    return callback;
  }

  // store io_manager
  // prevent deletion of io_manager before the tcp_client itself
  std::shared_ptr<multiplatform::io_manager> m_io_manager;

  // client socket
  multiplatform::tcp_socket m_socket;

  // whether the client is currently connected or not
  std::atomic<bool> m_is_connected = ATOMIC_VAR_INIT(false);

  std::queue<read_request> m_read_requests;   // read requests
  std::queue<write_request> m_write_requests; // write requests

  std::mutex m_read_requests_mutex;  // read requests thread safety
  std::mutex m_write_requests_mutex; // write requests thread safety

  disconnection_handler_t m_disconnection_handler;  // disconnection handler

  std::atomic<bool> m_debug = ATOMIC_VAR_INIT(false);
};

} // namespace multiplatform

#endif // MP_TCP_CLIENT_H_

