// This OnlineNnet2FeaturePipelineMod class is an addition to OnlineNnet2FeaturePipeline
// that corresponds to the case where OnlineFeatureInterface already exists.

#include "online-nnet2-feature-pipeline-mod.h"

namespace kaldi {

OnlineNnet2FeaturePipelineMod::OnlineNnet2FeaturePipelineMod(
    const OnlineNnet2FeaturePipelineInfo &info, std::shared_ptr<OnlineFeatureInterface> feature_if):
    info_(info), feature_if_(feature_if) {

  if(feature_if_ != nullptr){ /* nothing */ }
  else if(info_.feature_type == "mfcc"){   base_feature_ = new OnlineMfcc(info_.mfcc_opts); } 
  else if (info_.feature_type == "plp"){   base_feature_ = new OnlinePlp(info_.plp_opts); }
  else if (info_.feature_type == "fbank"){ base_feature_ = new OnlineFbank(info_.fbank_opts); }
  else{ KALDI_ERR << "Code error: invalid feature type " << info_.feature_type; }

  if(info_.add_pitch){
    pitch_ = new OnlinePitchFeature(info_.pitch_opts);
    pitch_feature_ = new OnlineProcessPitch(info_.pitch_process_opts, pitch_);
    if(feature_if_ != nullptr){
      feature_plus_optional_pitch_ = new OnlineAppendFeature(feature_if_.get(), pitch_feature_);
    }
    else{
      feature_plus_optional_pitch_ = new OnlineAppendFeature(base_feature_, pitch_feature_);
    }
  }
  else{
    pitch_ = NULL;
    pitch_feature_ = NULL;
    if(feature_if_ != nullptr){
      feature_plus_optional_pitch_ = feature_if.get();
    }
    else{
      feature_plus_optional_pitch_ = base_feature_;
    }
  }

  if(info_.use_ivectors){
    if(feature_if_ != nullptr){
      ivector_feature_ = new OnlineIvectorFeature(info_.ivector_extractor_info, feature_if_.get());
    }
    else{
      ivector_feature_ = new OnlineIvectorFeature(info_.ivector_extractor_info, base_feature_);
    }
    final_feature_ = new OnlineAppendFeature(feature_plus_optional_pitch_, ivector_feature_);
  }
  else{
    ivector_feature_ = NULL;
    final_feature_ = feature_plus_optional_pitch_;
  }
  dim_ = final_feature_->Dim();
}

int32 OnlineNnet2FeaturePipelineMod::Dim() const { return dim_; }

bool OnlineNnet2FeaturePipelineMod::IsLastFrame(int32 frame) const {
  return final_feature_->IsLastFrame(frame);
}

int32 OnlineNnet2FeaturePipelineMod::NumFramesReady() const {
  return final_feature_->NumFramesReady();
}

void OnlineNnet2FeaturePipelineMod::GetFrame(int32 frame,
                                          VectorBase<BaseFloat> *feat) {
  return final_feature_->GetFrame(frame, feat);
}

void OnlineNnet2FeaturePipelineMod::SetAdaptationState(
    const OnlineIvectorExtractorAdaptationState &adaptation_state) {
  if (info_.use_ivectors) {
    ivector_feature_->SetAdaptationState(adaptation_state);
  }
  // else silently do nothing, as there is nothing to do.
}

void OnlineNnet2FeaturePipelineMod::GetAdaptationState(
    OnlineIvectorExtractorAdaptationState *adaptation_state) const {
  if (info_.use_ivectors) {
    ivector_feature_->GetAdaptationState(adaptation_state);
  }
  // else silently do nothing, as there is nothing to do.
}


OnlineNnet2FeaturePipelineMod::~OnlineNnet2FeaturePipelineMod() {
  // Note: the delete command only deletes pointers that are non-NULL.  Not all
  // of the pointers below will be non-NULL.
  // Some of the online-feature pointers are just copies of other pointers,
  // and we do have to avoid deleting them in those cases.
  if(final_feature_ != feature_plus_optional_pitch_){
    delete final_feature_;
  }
  if(ivector_feature_){
    delete ivector_feature_;
  }
  if((feature_if_ == nullptr) && (feature_plus_optional_pitch_ != base_feature_)){
    delete feature_plus_optional_pitch_;
  }
  if(pitch_feature_){
    delete pitch_feature_;
  }
  if(pitch_){
    delete pitch_;
  }
  if(feature_if_ == nullptr){
    delete base_feature_;
  }
}

void OnlineNnet2FeaturePipelineMod::AcceptWaveform(
    BaseFloat sampling_rate,
    const VectorBase<BaseFloat> &waveform) {
  if(feature_if_ != nullptr){
    base_feature_->AcceptWaveform(sampling_rate, waveform);
    if(pitch_){
      pitch_->AcceptWaveform(sampling_rate, waveform);
    }
  }
}

void OnlineNnet2FeaturePipelineMod::InputFinished() {
  if(feature_if_ != nullptr){
    base_feature_->InputFinished();
    if(pitch_){
      pitch_->InputFinished();
    }
  }
}


}  // namespace kaldi
